# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import jsontableschema
from . import base
import copy

RESULTS = {
    'schema_001': {
        'id': 'schema_001',
        'name': 'Incorrect Headers',
        'msg': {
            'missing': 'Missing required column headings: {0}.',
            'misplaced': ('Misplaced column headings: {0}. Please respect the order'
                          ' set by the schema or change \"ignore_field_order\" to true.')
        },
        'help': '',
        'help_edit': ''
    },
    'schema_002': {
        'id': 'schema_002',
        'name': 'Incorrect Dimensions',
        'msg': 'The row dimensions do not match the header dimensions.',
        'help': '',
        'help_edit': ''
    },
    'schema_003': {
        'id': 'schema_003',
        'name': 'Incorrect Type',
        'msg': 'The value \"{0}\" in column \"{1}\" is not a valid {2}.',
        'help': '',
        'help_edit': ''
    },
    'schema_004': {
        'id': 'schema_004',
        'name': 'Required Field',
        'msg': 'Column \"{0}\" is a required field, but it contains no value.',
        'help': '',
        'help_edit': ''
    },
    'schema_005': {
        'id': 'schema_005',
        'name': 'Non-Required Field (Empty/Null)',
        'msg': 'Column \"{0}\" is a non-required field, and has a null value.',
        'help': '',
        'help_edit': ''
    },
    'schema_006': {
        'id': 'schema_006',
        'name': 'Unique Field',
        'msg': 'Column \"{0}\" is a unique field, yet the value \"{1}\" already exists.',
        'help': '',
        'help_edit': ''
    },
    'schema_007': {
        'id': 'schema_007',
        'name': 'Incorrect Type Extra Field',
        'msg': ('Column \"{0}\" is an extra field.'
                'However, the value \"{1}\" is not a valid {2}.'),
        'help': '',
        'help_edit': ''
    },
    'schema_008': {
        'id': 'schema_008',
        'name': 'Extra Field (Empty/Null)',
        'msg': 'Column \"{0}\" is an extra field and it has a null value.',
        'help': '',
        'help_edit': ''
    }
}


class SchemaProcessor(base.Processor):

    """Process data against a JSON Table Schema."""

    name = 'schema'
    RESULT_TYPES = RESULTS


    def __init__(self, fail_fast=False, report_limit=1000,
                 row_limit=30000, schema=None, ignore_field_order=True,
                 report_stream=None, report=None, result_level='error',
                 infer_schema=False, process_extra_fields=False,
                 case_insensitive_headers=False, **kwargs):

        super(SchemaProcessor, self).__init__(
            fail_fast=fail_fast, report_limit=report_limit,
            row_limit=row_limit, report_stream=report_stream, report=report,
            result_level=result_level)

        self.infer_schema = infer_schema
        self.case_insensitive_headers = case_insensitive_headers
        self.ignore_field_order = ignore_field_order
        self.process_extra_fields = process_extra_fields

        if not schema:
            self.schema = None
        else:
            self.schema = self.schema_model(schema)

        self._uniques = {}

    def schema_model(self, schema):
        try:
            model = jsontableschema.model.SchemaModel(schema, self.case_insensitive_headers)
        except (jsontableschema.exceptions.InvalidJSONError, jsontableschema.exceptions.InvalidSchemaError) as e:
            raise e

        return model

    def pre_run(self, data_table):

        sample_values = data_table.get_sample(300)

        if (self.schema is None) and self.infer_schema:
            self.schema = self.schema_model(jsontableschema.infer(data_table.headers, sample_values))

        if self.schema and self.process_extra_fields:
            self.extra_fields = (set(data_table.headers)).difference(set(self.schema.headers))
            infered_schema = jsontableschema.infer(data_table.headers, sample_values)
            complete_schema_dict = self.schema._to_python()

            for field in infered_schema['fields']:
                if field['name'] in self.extra_fields:
                    complete_schema_dict['fields'].append(copy.deepcopy(field))

            self.schema = self.schema_model(complete_schema_dict)

        return True, data_table

    def run_header(self, headers, header_index=0):

        valid = True

        if self.case_insensitive_headers:
            headers = [name.lower() for name in headers]

        if self.schema:
            required_headers = self.schema.required_headers
            required_headers_set = set(required_headers)
            headers_set = set(headers)

            if self.ignore_field_order:
                if not headers_set.issuperset(required_headers_set):
                    missing_set = required_headers_set.difference(headers_set)
                    valid = False
                    _type = RESULTS['schema_001']
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg']['missing'].format(', '.join(missing_set)),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

            else:
                req_headers_len = len(required_headers)
                target_headers = headers[:req_headers_len]

                if not (target_headers == required_headers):
                    missing_set = required_headers_set.difference(headers_set)
                    misplaced_set = [th for th, rh in zip(target_headers, required_headers)
                                     if th != rh]
                    valid = False
                    _type = RESULTS['schema_001']
                    msg = lambda set, key: _type['msg'][key].format(', '.join(set)) if set else ''
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        ' '.join([msg(missing_set, 'missing'),
                                 msg(misplaced_set, 'misplaced')]),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME,
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

        return valid, headers

    def run_row(self, headers, index, row):

        valid = True
        row_name = self.get_row_id(headers, row)

        if self.schema:
            if not (len(headers) == len(row)):

                valid = False
                _type = RESULTS['schema_002']
                entry = self.make_entry(
                    self.name,
                    self.RESULT_CATEGORY_ROW,
                    self.RESULT_LEVEL_ERROR,
                    _type['msg'],
                    _type['id'],
                    _type['name'],
                    row,
                    index,
                    row_name,
                )

                self.report.write(entry)
                if self.fail_fast:
                    return valid, headers, index, row

            else:
                for column_name, column_value in zip(headers, row):

                    # handle case where column_name not even in schema
                    if not self.schema.has_field(column_name):
                        pass

                    # we know the field is in the schema
                    else:
                        # CONSTRAINTS
                        constraints = self.schema.get_constraints(column_name)

                        if constraints is not None and \
                           constraints.get('required', False) is True and \
                           (column_value in self.schema.NULL_VALUES):

                            valid = False
                            _type = RESULTS['schema_004']
                            entry = self.make_entry(
                                self.name,
                                self.RESULT_CATEGORY_ROW,
                                self.RESULT_LEVEL_ERROR,
                                _type['msg'].format(column_name),
                                _type['id'],
                                _type['name'],
                                row,
                                index,
                                row_name,
                                headers.index(column_name),
                                column_name
                            )

                            self.report.write(entry)
                            if self.fail_fast:
                                return valid, headers, index, row

                        if constraints is not None and \
                           constraints.get('required', False) is False and \
                           (column_value in self.schema.NULL_VALUES) and \
                           self.result_level == self.RESULT_LEVEL_INFO:
                            # add info result
                            _type = RESULTS['schema_005']
                            entry = self.make_entry(
                                self.name,
                                self.RESULT_CATEGORY_ROW,
                                self.RESULT_LEVEL_INFO,
                                _type['msg'].format(column_name),
                                _type['id'],
                                _type['name'],
                                row,
                                index,
                                row_name,
                                headers.index(column_name),
                                column_name
                            )

                            self.report.write(entry)
                            if self.fail_fast:
                                return valid, headers, index, row

                        if self.process_extra_fields and \
                           column_name in self.extra_fields and \
                           column_value in self.schema.NULL_VALUES and \
                           self.result_level == self.RESULT_LEVEL_INFO:

                            _type = RESULTS['schema_008']
                            entry = self.make_entry(
                                self.name,
                                self.RESULT_CATEGORY_ROW,
                                self.RESULT_LEVEL_INFO,
                                _type['msg'].format(column_name),
                                _type['id'],
                                _type['name'],
                                row,
                                index,
                                row_name,
                                headers.index(column_name),
                                column_name
                            )

                            self.report.write(entry)
                            if self.fail_fast:
                                return valid, headers, index, row

                        if constraints is not None and \
                            constraints.get('unique') is True:

                            if not self._uniques.get(column_name):
                                self._uniques[column_name] = set([column_value])

                            elif column_value in self._uniques[column_name]:
                                _type = RESULTS['schema_006']
                                entry = self.make_entry(
                                    self.name,
                                    self.RESULT_CATEGORY_ROW,
                                    self.RESULT_LEVEL_ERROR,
                                    _type['msg'].format(column_name, column_value),
                                    _type['id'],
                                    _type['name'],
                                    row,
                                    index,
                                    row_name,
                                    headers.index(column_name),
                                    column_name
                                )

                                self.report.write(entry)
                                if self.fail_fast:
                                    return valid, headers, index, row

                            else:
                                self._uniques[column_name].add(column_value)

                        # TODO: check constraints.min* and constraints.max*
                        # check type and format
                        try:
                            schema_cast = self.schema.cast(column_name,
                                                           column_value)
                        except (jsontableschema.exceptions.InvalidCastError,
                                jsontableschema.exceptions.ConstraintError):
                            schema_cast = False

                        if schema_cast is False:
                            expected_type = self.schema.get_type(column_name).name.title()

                            if self.process_extra_fields and \
                               column_name in self.extra_fields and \
                               self.result_level == self.RESULT_LEVEL_INFO:

                                _type = RESULTS['schema_007']
                                entry = self.make_entry(
                                    self.name,
                                    self.RESULT_CATEGORY_ROW,
                                    self.RESULT_LEVEL_INFO,
                                    _type['msg'].format(column_name, column_value,
                                                        expected_type),
                                    _type['id'],
                                    _type['name'],
                                    row,
                                    index,
                                    row_name,
                                    headers.index(column_name),
                                    column_name
                                )

                                self.report.write(entry)
                                if self.fail_fast:
                                    return valid, headers, index, row

                            else:

                                valid = False
                                _type = RESULTS['schema_003']
                                entry = self.make_entry(
                                    self.name,
                                    self.RESULT_CATEGORY_ROW,
                                    self.RESULT_LEVEL_ERROR,
                                    _type['msg'].format(column_value, column_name,
                                                        expected_type),
                                    _type['id'],
                                    _type['name'],
                                    row,
                                    index,
                                    row_name,
                                    headers.index(column_name),
                                    column_name
                                )

                                self.report.write(entry)
                                if self.fail_fast:
                                    return valid, headers, index, row

        return valid, headers, index, row



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class InvalidSpec(Exception):
    pass


class PipelineBuildError(Exception):

    def __init__(self, msg=None):
        self.msg = msg or 'Could not build the pipeline.'


class ProcessorBuildError(Exception):

    def __init__(self, msg=None):
        self.msg = msg or 'Could not build the processor.'


class DataSourceHTTPError(Exception):

    def __init__(self, msg=None, status=None, name=None):

        from goodtables import compat

        self.status = status
        self.msg = msg or ('The data source server responded with an error: '
                           '{0}.').format(compat.responses.get(self.status, ''))
        self.name = name or '{0} Error'.format(compat.responses.get(self.status, 'http'))

        
class DataSourceFormatUnsupportedError(Exception):

    def __init__(self, msg=None, file_format=None, name=None):
        self.file_format = file_format or 'unsupported'
        self.msg = msg or 'The data source format is {0}.'.format(self.file_format.upper())
        self.name = name or 'Data Source {0} Error'.format(self.file_format)


class DataSourceDecodeError(Exception):

    def __init__(self, msg=None, name=None):
        self.msg = msg or ('The data source cannot be decoded using the '
                           'declared or detected encoding.')
        self.name = name or 'Data Source Decode Error'

    
class DataSourceMalformatedError(Exception):

    def __init__(self, msg=None, file_format=None, name=None):
        self.file_format = file_format or 'tabular file'
        self.msg = 'The data source is an invalid {0}: {1}.'.format(self.file_format, msg)
        self.name = name or 'Invalid {0} Error'.format(self.file_format).title()


class InvalidHandlerError(Exception):

    def __init__(self, msg=None):
        self.msg = msg or 'The passed handler is not valid.'

class InvalidPipelineOptions(ValueError):

    def __init__(self, msg=None):
        self.msg = msg or 'Some of the options are not valid.'



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .pipeline import Pipeline
from .batch import Batch


__all__ = ['Pipeline', 'Batch']



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import io
import importlib
import tellme
import inspect
from ..utilities import csv_dialect, helpers
from .. import datatable
from .. import exceptions
from ..processors import base


class Pipeline(object):

    """Validate a (tabular) data source through a validation pipeline.

    Args:
    * processors: A list of processor names to process `data_source`
        * Each name can be a 'shortname' for the default processors
            * e.g., ['structure', 'schema']
        * Each name can be a string path to a processor
            * e.g., ['custompackage.CustomValidator', 'schema']
            * Custom processors must implement the Validator API
    * data: A buffer, filepath, string or URL to the table data
    * format: The format of `data_source`. 'csv' or 'json'
    * dialect: A buffer, filepath, string or URL to a CSV dialect spec
    * options: a dict configuration object for the pipeline
        * Each processor has its options nested under its 'shortname'
        * Custom processors have options nested under cls.__name__.lower()
        * e.g.:
            {'structure': {#options}, 'customprocessor': {#options}}

    Returns:
        A tuple of `valid, report`, where `valid` is a boolean expressing
        validity according to the whole pipeline, and report is a dict
        of the entire pipeline report where each top-level key matches
        a processor in the pipeline.

    """

    ROW_LIMIT_MAX = 30000
    REPORT_LIMIT_MAX = 1000
    DIALECT_FILENAME = 'dialect.json'
    DATA_FORMATS = ('csv', 'excel', 'json')

    def __init__(self, data, processors=None, dialect=None, format='csv',
                 transform=True, encoding=None, decode_strategy='replace',
                 options=None, fail_fast=False, row_limit=20000,
                 report_limit=1000, report_stream=None, header_index=0,
                 break_on_invalid_processor=True, post_task=None,
                 report_type='basic'):

        if data is None:
            _msg = '`data` must be a filepath, url or stream.'
            raise exceptions.PipelineBuildError(_msg)

        self.openfiles = []
        self.data_source = data
        self.processors = processors or helpers.DEFAULT_PIPELINE
        self.dialect = self.get_dialect(dialect)
        self.format = format
        self.encoding = encoding
        self.decode_strategy = decode_strategy
        self.options = options or {}
        self.transform = transform
        self.fail_fast = fail_fast
        self.row_limit = self.get_row_limit(row_limit)
        self.report_limit = self.get_report_limit(report_limit)
        self.report_stream = report_stream
        self.header_index = header_index
        self.break_on_invalid_processor = break_on_invalid_processor

        helpers.validate_handler(post_task)

        if report_type == 'grouped':
            self.report_post_task = helpers.grouped_report
        else:
            self.report_post_task = helpers.basic_report

        self.post_task = post_task

        if self.report_stream:
            report_stream_tests = [isinstance(self.report_stream, io.TextIOBase),
                                   self.report_stream.writable(),
                                   self.report_stream.seekable()]

            if not all(report_stream_tests):
                _msg = '`report_stream` must be a seekable and writable text stream.'
                raise exceptions.PipelineBuildError(_msg)

            report_backend = 'client'
        else:
            report_backend = 'yaml'

        report_options = {
            'schema': helpers.report_schema,
            'backend': report_backend,
            'client_stream': self.report_stream,
            'limit': self.report_limit,
            'post_task': self.report_post_task
        }

        self.report = tellme.Report('Pipeline', **report_options)

        self.pipeline = self.get_pipeline()

        try:
            self.data = datatable.DataTable(self.data_source, format=self.format,
                                            encoding=encoding,
                                            decode_strategy=decode_strategy,
                                            header_index=self.header_index)
            self.openfiles.extend(self.data.openfiles)
            
        except datatable.DataTable.RAISES:
            self.data = self.data_source
       

    def get_pipeline(self):
        """Get the pipeline for this instance."""

        self.validate_processors()
        self.validate_options()
        pipeline = []
        for name in self.processors:
            _class = self.resolve_processor(name)
            options = self.options.get(_class.name, {})
            # override options with those we accept directly from the pipeline
            options['row_limit'] = self.row_limit
            options['report_limit'] = self.report_limit
            options['report_stream'] = self.report_stream
            options['fail_fast'] = self.fail_fast
            options['report'] = self.report
            options['transform'] = self.transform
            options['header_index'] = self.header_index

            try:
                instance = _class(**options)
                pipeline.append(instance)
            except _class.RAISES as e:
                raise e

        return pipeline

    def validate_processors(self):
        """Validates the processors parameter"""

        if not isinstance(self.processors, (list, tuple, set)):
            msg = 'The \'processors\' argument must be a list, tuple or set.'
            raise exceptions.InvalidPipelineOptions(msg)

        for processor_name in self.processors:
            self.resolve_processor(processor_name)

    def validate_options(self):
        """Validates the options parameter."""

        if not isinstance(self.options, dict):
            msg = 'Pipeline \'options\' argument must be a dict.'
            raise exceptions.InvalidPipelineOptions(msg)

        if not set(self.options.keys()).issubset(set(self.processors)):
            unknown_opts = set(self.options.keys()).difference(set(self.processors))
            msg = ('Option(s) \'{0}\' don\'t correspond to any required '
                   'processor.').format(','.join(unknown_opts))
            raise exceptions.InvalidPipelineOptions(msg)

        for processor_name in self.processors:
            passed_opts = self.options.get(processor_name, {})
            processor = self.resolve_processor(processor_name)
            processor_args, _, _, _, = inspect.getargspec(processor.__init__)
            base_args, _, _, _, = inspect.getargspec(base.Processor.__init__)
            available_opts = processor_args[1::]
            available_opts.extend(base_args[1::])
            invalid_opts = [option for option in passed_opts.keys()
                                   if option not in available_opts]
            if invalid_opts:
                msg = ('Option(s) \'{0}\' are not valid arguments for {1} '
                       'processor.').format(','.join(invalid_opts), processor_name)
                raise exceptions.InvalidPipelineOptions(msg)

        return None

    def get_dialect(self, dialect_source):
        """Get a CSV dialect instance for this data."""

        if dialect_source:
            _valid, rv = csv_dialect.validate(dialect_source)
            if not _valid:
                raise exceptions.InvalidSpec
            return rv
        else:
            return None

    def get_row_limit(self, passed_limit):
        """Return the row limit, locked to an upper limit."""

        if passed_limit > self.ROW_LIMIT_MAX:
            return self.ROW_LIMIT_MAX
        else:
            return passed_limit

    def get_report_limit(self, passed_limit):
        """Return the report_limit, locked to an upper limit."""

        if passed_limit > self.REPORT_LIMIT_MAX:
            return self.REPORT_LIMIT_MAX
        else:
            return passed_limit

    def resolve_processor(self, processor_name):
        """Return a processor class."""

        if processor_name in helpers.builtin_processors():
            processor_class = helpers.builtin_processors()[processor_name]
        else:
            # resolve a custom processor
            _module, _class = processor_name.rsplit('.', 1)
            try:
                processor_class = getattr(importlib.import_module(_module),
                                          _class)
            except ImportError:
                msg = ('The processsor \'{0}\' could not be resolved due to an '
                       'Import Error.').format(processor_name)
                raise exceptions.InvalidPipelineOptions(msg)

        return processor_class

    def register_processor(self, processor_name, options=None, position=None):
        """Register a processor on the pipeline."""

        processor_class = self.resolve_processor(processor_name)
        options = options or {}
        processor = processor_class(**options)

        if position is None:
            self.pipeline.append(processor)
        else:
            self.pipeline.insert(position, processor)

    def run(self):
        """Run the pipeline."""

        def _run_valid(process_valid, run_valid):
            """Set/maintain the valid state of the run."""
            if not process_valid and run_valid:
                return False
            return run_valid

        valid = True

        for processor in self.pipeline:
            
            if isinstance(self.data, datatable.DataTable):
                _valid, _, self.data = processor.run(self.data, is_table=True,
                    encoding=self.encoding, decode_strategy=self.decode_strategy)
            else:
                _valid, _, self.data = processor.run(self.data_source, is_table=False,
                                            decode_strategy=self.decode_strategy,
                                            encoding=self.encoding, format=self.format)
            valid = _run_valid(_valid, valid)
            
            # if a validator returns invalid, we stop the pipeline,
            # unless break_on_invalid_processor is False
            if not valid and self.break_on_invalid_processor:
                break

            if self.data:
                self.data.replay()

        self.set_report_meta()

        if self.post_task:
            # TODO: handle what happens in here
            self.post_task(self)

        return valid, self.report

    def set_report_meta(self):
        """Set information and statistics for this run on report['meta']."""

        if self.data:
            self.report.meta['row_count'] = self.pipeline[0].row_count or 1
            self.report.meta['header_index'] = self.header_index
            self.report.meta['headers'] = self.data.headers
            self.report.meta['encoding'] = self.data.encoding
        else: 
            self.report.meta['row_count'] = 0
            self.report.meta['header_index'] = 0
            self.report.meta['headers'] = []
           



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
import json
import inspect
import requests
from .. import compat
from .. import exceptions


def builtin_processors():
    """Return dict of public builtin processors. Avoids circular import."""
    from .. import processors
    return {
        processors.StructureProcessor.name: processors.StructureProcessor,
        processors.SchemaProcessor.name: processors.SchemaProcessor
    }


DEFAULT_PIPELINE = ('structure',)
REMOTE_SCHEMES = ('http', 'https', 'ftp', 'ftps')


# a schema for the reporter.Report() instances used by validators
report_schema = {
    'processor': {'type': compat.str},
    'result_category': {'type': compat.str},
    'result_level': {'type': compat.str},
    'result_message': {'type': compat.str},
    'result_id': {'type': compat.str},
    'result_name': {'type': compat.str},
    'result_context': {'type': list},
    'row_index': {'type': (int, type(None))},
    'row_name': {'type': compat.str},
    'column_index': {'type': (int, type(None))},
    'column_name': {'type': compat.str}
}


def get_report_result_types():
    """Return a list of all builtin result types."""

    result_types = []

    for processor in builtin_processors().values():
        result_types.extend([r[1] for r in
                             tuple(processor.RESULT_TYPES.items())])

    return result_types


def basic_report(report):
    """Run meta statistics."""

    report = set_meta_stats(report)

    return report


def grouped_report(report):
    """Run meta statistics and group the results by row."""

    report = set_meta_stats(report)
    report = group_results(report)

    return report


def set_meta_stats(report):

    """Set run statistics on the report meta object."""

    def _bad_type_percent(column_name, results, row_count):
        """Return a percentage for amount of data with bad type in a column."""

        match_count = len([r for r in results if
                           r['result_id'] == 'schema_003' and
                           r['column_name'] == column_name])

        return int((match_count/row_count) * 100)

    bad_row_count = len(set([r['row_index'] for r in report['results'] if
                             r['result_category'] == 'row' and
                             r['result_level'] == 'error' and
                             r['row_index'] is not None]))

    bad_column_count = len(set([r['column_index'] for r in report['results'] if
                                r['result_category'] == 'row' and
                                r['result_level'] == 'error' and
                                r['column_index'] is not None]))

    columns = [{'name': header, 'index': index,
                'bad_type_percent': _bad_type_percent(header, report['results'],
                                                      report['meta']['row_count'])}
               for index, header in enumerate(report['meta']['headers'])]

    report['meta'].update({
        'bad_row_count': bad_row_count,
        'bad_column_count': bad_column_count,
        'columns': columns
    })

    return report


def group_results(report):

    """Group report results by row."""

    # If the report is a base error
    base_reports = [result for result in report['results']
                    if result['processor'] == 'base']
    if any(base_reports):
        group_report = {'':{'result_context': [], 'results': base_reports }}
        report['results']= [ group_report ]
        return report

    _rows = set([r['row_index'] for r in report['results']
                 if r['row_index'] is not None])

    def make_groups(results, rows):
        groups = {}

        for row in rows:
            groups.update({
                row: {
                    'row_index': row,
                    'results': []
                }
            })

        for index, result in enumerate(results):
            if result['row_index'] is not None:

                # set the result context on the group
                if not groups[result['row_index']].get('result_context'):
                    groups[result['row_index']]['result_context'] = result['result_context']

                groups[result['row_index']]['results'].append(result)

                # remove stuff we do not require per result
                del result['result_context']
                del result['row_index']

        return groups

    report['results'] = sorted([{k: v} for k, v in make_groups(report['results'], _rows).items()],
                               key=lambda result: list(result.keys())[0])
    return report


def load_json_source(source):

    """Load a JSON source, from string, URL or buffer,  into a Python type."""

    if source is None:
        return None

    elif isinstance(source, (dict, list)):
        # the source has already been loaded
        return source

    elif compat.urlparse(source).scheme in REMOTE_SCHEMES:
        valid_url = make_valid_url(source)
        return json.loads(requests.get(valid_url).text)

    elif isinstance(source, compat.str) and not os.path.exists(source):
        return json.loads(source)

    else:
        with io.open(source, encoding='utf-8') as stream:
            source = json.load(io.open(source, encoding='utf-8'))
        return source


def validate_handler(handler, argcount=1):

    """Validate a handler.

    Args:
        * `handler`: the supposed handler, which is a function
        * `argcount`: the count of args the handler should take
        (not including self if handler is a method)
    """

    if handler is None:
        return True
    else:
        if not callable(handler):
            raise exceptions.InvalidHandlerError

        if inspect.ismethod(handler):
            # extra arg for self
            argcount += 1

        spec = inspect.getargspec(handler)
        if not len(spec[0]) == argcount:
            raise exceptions.InvalidHandlerError

def make_valid_url(url):
    """Make sure url doesn't contain unsupported characters

    Args:
        * `url`: a url string
    """
    if '/+/http' in url:
        glue = '/+/'
        quoted = [make_valid_url(unquoted) for unquoted in url.split(glue)]
        return (glue).join(quoted)

    scheme, netloc, path, query, fragment = compat.urlsplit(url)
    quoted_path = compat.quote(path.encode('utf-8'))
    quoted_query = compat.quote_plus(query.encode('utf-8'))
    new_url_tuple = (scheme, netloc, quoted_path, quoted_query, fragment)
    quoted_url = compat.urlunsplit(new_url_tuple)
    return quoted_url
